      MODULE FillMatrices

! This module fills the Right and Left hand side matrices 
!  of the normal equation using 
! grid and base functions defined in the previous module 
! and weighting procedure and integration rules defined here.

      use constants
!
      use defbasegrid,ONLY:cdim,gdim,gn1,gn2,gn3,sgn,cell,
     &  gridMask,gridMask2,gridMask3,
     &  baski,baski2,baski3,gridvol,gridvol2,gridvol3,posi,
     &  posi2,posi3,invsuski,reqaccuracy,grs,grs2,grs3
!
      use incidentia,ONLY:IncidentFields
!
      use matrixia,ONLY:matsqr3,matmul3,matmul3cc,matmul3a

      IMPLICIT NONE


!       FillTheMatrices

      CONTAINS 



! ****************************************************************************
	  SUBROUTINE FillTheMatrices(A,e0,dim,Orientations,pp)
! ****************************************************************************
!
! Here the innerproducts  e0=<w_i|E_0> and A(i,j)=<w_i|w_j>
!
! with      |w_j> = |f_j> - |Gs*f_j>
! and       <a|b> =  integ(v,conj(a)*b) dv  are computed.
!
!
! =>    e0 =
!
!           <f_i|E_0> - <f_i|s'G'|E_0> 
! and
!       Aij =
! 
!       <f_i|f_j> -<f_i|s'G'|f_j>-<f_i|Gs|f_j> + <f_i|s'G'Gs|f_j>
! with
!
!       <f_i|E_0> - <f_i|s'G'|k>  <k|E_0> and 
!
!       <f_i|f_j> -<f_i|s'G'|k><k|f_j>-<f_i|k><k|Gs|f_j> +
!
!                             <f_i|s'G'|k><k|k'><k'|Gs|f_j>

! SPEED CRITICAL SUBROUTINE
! Any help in improving the speed and accuracy is highly appreciated.

	 USE Timing,ONLY:CPUtime

	 INTEGER,INTENT(in) ::dim,Orientations,pp
	 COMPLEX*16,INTENT(out),dimension(dim,dim)::A
	 COMPLEX*16,INTENT(out),
     &     dimension(dim,Orientations,pp)::e0
!
	 INTEGER c0
	 REAL*4 vol,gvol
	 REAL*4,DIMENSION(3):: r0
!
	 COMPLEX*16,DIMENSION(3,dim):: bG3,bG2,bG1,bGV1,bGV2
	 COMPLEX*16,DIMENSION(dim,3):: cG1,cG2,cG3
	 COMPLEX*16,DIMENSION(3,Orientations):: Ei1,Ei2
	 COMPLEX*16,DIMENSION(:,:),
     &       ALLOCATABLE:: casis,basis,A1,A2,e1(:,:,:)
	 LOGICAL,DIMENSION(0:cdim):: GreenMask,GreenMask2,GreenMask3
!??????
!        COMPLEX*8,DIMENSION(:,:,:,:,:,:),
!???????
	 COMPLEX*16,DIMENSION(:,:,:,:,:,:),
     &       ALLOCATABLE::SingularValue3
	 INTEGER g1,g2,g3,gg1,gg2,gg3,ggg1,ggg2,ggg3,grid0,gridpointspcell
	 LOGICAL:: LeastSquaring=.TRUE.  ! otherwise Galerkin

	  e0=0.0
	  A=0.0 
	  vol=0.0
	  gridpointspcell=gn1*gn2*gn3
!
	  ALLOCATE(SingularValue3(3,3,gridpointspcell,gridpointspcell,
     &      gridpointspcell,1:gdim))
!
! compute singularities only once !!!!
!
	  CALL DoSingularities3(cdim,gdim,gridpointspcell,
     &     SingularValue3,grs) 
!
	  WRITE(*,'(A,I4,F10.1)') '^',sgn**3,CPUtime()
!
! cell-loop   cell-nr.:c0=1..cdim
!       
       IF (dim>=1000) THEN
!
	  DO c0=0,cdim
!
	    grid0=cell(c0)%grid
	    GreenMask=.false.
	    ALLOCATE(e1(cell(c0)%exporder,Orientations,pp));e1=0.0
	    ALLOCATE(A1(cell(c0)%exporder,dim));A1=0.0
	    ALLOCATE(A2(cell(c0)%exporder,cell(c0)%exporder));A2=0.0
	    ALLOCATE(casis(cell(c0)%exporder,3),
     &                basis(3,cell(c0)%exporder))
!
! gridpoint-loop
!
	    DO ggg1=1,gn3;  DO ggg2=1,gn3;  DO ggg3=1,gn3
!
	     IF (gridMask3(ggg1,ggg2,ggg3,grid0)) THEN
	      r0=cell(c0)%center+posi3(ggg1,ggg2,ggg3,:,grid0)
	      GreenMask3=FarMask(r0,c0,3)
	      IF (Any(GreenMask3)) THEN
		bG3= Greenint(dim,c0,r0,ggg1,ggg2,ggg3,GreenMask3,3,gn3)
                cG3=Transpose(Conjg(bG3))
		IF (LeastSquaring) THEN
!
		  CALL IncidentFields(Ei1,r0,Orientations,1)
		  CALL IncidentFields(Ei2,r0,Orientations,2)
!
! to circumvent a segment violation bug in DEC f90 1.0
!   matsqr3 faster than matmul(transp etc. on risc6000
!
 		     CALL matsqr3(A,bG3,gridvol3(ggg1,ggg2,ggg3,grid0))  
!
!		  e0(:,:,1)=e0(:,:,1)-
!    &                   matmul3(cG3,Ei1*gridvol3(ggg1,ggg2,ggg3,grid0))
!
                CALL matmul3a(e0(:,:,1),cG3,Ei1*
     &                   (-gridvol3(ggg1,ggg2,ggg3,grid0)))
!
!		  e0(:,:,2)=e0(:,:,2)-
!    &                   matmul3(cG3,Ei2*gridvol3(ggg1,ggg2,ggg3,grid0)
!
                CALL matmul3a(e0(:,:,2),cG3,Ei2*
     &                   (-gridvol3(ggg1,ggg2,ggg3,grid0)))
		ENDIF
	      ELSE
		  bG3=0.0;cG3=0.0
	      ENDIF
!
	      bGV2=0.0
	      DO gg1=(ggg1-1)*gn2+1,ggg1*gn2
              DO gg2=(ggg2-1)*gn2+1,ggg2*gn2
              DO gg3=(ggg3-1)*gn2+1,ggg3*gn2
	       IF (gridMask2(gg1,gg2,gg3,grid0)) THEN
		r0=cell(c0)%center+posi2(gg1,gg2,gg3,:,grid0)
		GreenMask2=FarMask(r0,c0,2).AND..NOT.GreenMask3
		IF (Any(GreenMask2)) THEN
		  bG2=Greenint(dim,c0,r0,gg1,gg2,gg3,GreenMask2,2,gn2*gn3)
                  cG2=Transpose(Conjg(bG2))
		  bGV2=bGV2+bG2*gridvol2(gg1,gg2,gg3,grid0)
		  IF (LeastSquaring) THEN
!
		    CALL IncidentFields(Ei1,r0,Orientations,1)
		    CALL IncidentFields(Ei2,r0,Orientations,2)
!
! to circumvent a segment violation bug in DEC f90 1.0
!
 		       CALL matsqr3(A,bG2,gridvol2(gg1,gg2,gg3,grid0))  
!
!		    e0(:,:,1)=e0(:,:,1)-matmul3(cG2,Ei1)*
!    &                   gridvol2(gg1,gg2,gg3,grid0)
!
		CALL matmul3a(e0(:,:,1),cG2,Ei1*
     &                (-gridvol2(gg1,gg2,gg3,grid0)))
!
!		    e0(:,:,2)=e0(:,:,2)-matmul3(cG2,Ei2)*
!    &                   gridvol2(gg1,gg2,gg3,grid0)
!
		CALL matmul3a(e0(:,:,2),cG2,Ei2*
     &                (-gridvol2(gg1,gg2,gg3,grid0)))
		  ENDIF
		ELSE
		  bG2=0.0;cG2=0.0
		ENDIF
		bGV1=0.0
		DO g1=(gg1-1)*gn1+1,gg1*gn1
                DO g2=(gg2-1)*gn1+1,gg2*gn1
                DO g3=(gg3-1)*gn1+1,gg3*gn1 
		  IF (gridMask(g1,g2,g3,grid0)) THEN
		    GreenMask=(.NOT.GreenMask3).AND..NOT.GreenMask2
		    gvol=gridvol(g1,g2,g3,grid0)
		    vol=vol+gvol
		    r0=cell(c0)%center+posi(g1,g2,g3,:,grid0)
		    bG1=Greenint(dim,c0,r0,g1,g2,g3,GreenMask,1,gn1*gn2*gn3)
                    cG1=Transpose(Conjg(bG1))
		    bGV1=bGV1+bG1*gvol
		    basis=invsuski(g1,g2,g3,grid0)*
     &              baski(g1,g2,g3,:,1:cell(c0)%exporder,grid0)
     &              -matmul3(SingularValue3(:,:,g1,g2,g3,grid0),
     &                baski(g1,g2,g3,:,1:cell(c0)%exporder,grid0))
		    IF (LeastSquaring) THEN
			casis=TRANSPOSE(conjg(basis))
		    ELSE
			casis=Transpose(Conjg(invsuski(g1,g2,g3,grid0)*
     &                  baski(g1,g2,g3,:,1:cell(c0)%exporder,grid0)))
		    ENDIF
!
		    CALL IncidentFields(Ei1,r0,Orientations,1)
		    CALL IncidentFields(Ei2,r0,Orientations,2)
!
!       	    e1(:,:,1) = e1(:,:,1)+matmul3(casis,Ei1)
!       	    e1(:,:,2) = e1(:,:,2)+matmul3(casis,Ei2)
!
		    CALL matmul3a(e1(:,:,1),casis,Ei1)
		    CALL matmul3a(e1(:,:,2),casis,Ei2)
		    IF (LeastSquaring) THEN
		       e0(:,:,1) = e0(:,:,1)-matmul3(cG1,Ei1)*gvol
		       e0(:,:,2) = e0(:,:,2)-matmul3(cG1,Ei2)*gvol
!-------------------------------------------------------------------------
!		       IF (dim>=1000) THEN    
! to circumvent a segment violation bug in DEC f90 1.0
!
 			 CALL matsqr3(A,bG1,gvol) 
!
!---------------------------------------------------------------------------
		    ENDIF
		    A1=A1-matmul3(casis,bG1+bG2+bG3)
		    A2=A2+matmul3(casis,basis)/gvol
		  ENDIF
		ENDDO;ENDDO;ENDDO
!        IF (Any(GreenMask2)) A(:,:)=A(:,:)+(matmul3(cG2,bGV1)+
!    &          conjg(transpose(matmul3(cG2,bGV1))))
		IF (Any(GreenMask2).AND.LeastSquaring)    
!
     &                   CALL matmul3cc(A,cG2,bGV1)
		bGV2=bGV2+bGV1
!
	       ENDIF
	      ENDDO;ENDDO;ENDDO
!       IF (Any(GreenMask3)) A(:,:)=A(:,:)+(matmul3(cG3,bGV2)+
!    &          conjg(transpose(matmul3(cG3,bGV2))))
	       IF (Any(GreenMask3).AND.LeastSquaring) 
!
     &                     CALL matmul3cc(A,cG3,bGV2)
!
	      ENDIF
	     ENDDO;ENDDO;ENDDO
!
! end of gridpoint-loop
!
	    WRITE(*,'(A,I4,F10.1)') '#',c0,CPUtime()
	    e0(cell(c0)%first:cell(c0)%last,:,1) =
     &          e0(cell(c0)%first:cell(c0)%last,:,1)+e1(:,:,1)
	    e0(cell(c0)%first:cell(c0)%last,:,2) =
     &          e0(cell(c0)%first:cell(c0)%last,:,2)+e1(:,:,2)
	    A(cell(c0)%first:cell(c0)%last,:)=
     &          A(cell(c0)%first:cell(c0)%last,:)+A1
	    IF (LeastSquaring) A(:,cell(c0)%first:cell(c0)%last)=
     &        A(:,cell(c0)%first:cell(c0)%last)+conjg(transpose(A1))
           A(cell(c0)%first:cell(c0)%last,cell(c0)%first:cell(c0)%last)=
     &     A(cell(c0)%first:cell(c0)%last,
     &     cell(c0)%first:cell(c0)%last)+A2
	    DEALLOCATE(casis,basis)
	    DEALLOCATE(A2)
	    DEALLOCATE(A1)
	    DEALLOCATE(e1)
	 ENDDO
      ELSE
	  DO c0=0,cdim
!
	    grid0=cell(c0)%grid
	    GreenMask=.false.
	    ALLOCATE(e1(cell(c0)%exporder,Orientations,pp));e1=0.0
	    ALLOCATE(A1(cell(c0)%exporder,dim));A1=0.0
	    ALLOCATE(A2(cell(c0)%exporder,cell(c0)%exporder));A2=0.0
	    ALLOCATE(casis(cell(c0)%exporder,3),
     &                basis(3,cell(c0)%exporder))
!
! gridpoint-loop
!
	    DO ggg1=1,gn3;  DO ggg2=1,gn3;  DO ggg3=1,gn3
!
	     IF (gridMask3(ggg1,ggg2,ggg3,grid0)) THEN
	      r0=cell(c0)%center+posi3(ggg1,ggg2,ggg3,:,grid0)
	      GreenMask3=FarMask(r0,c0,3)
	      IF (Any(GreenMask3)) THEN
		bG3= Greenint(dim,c0,r0,ggg1,ggg2,ggg3,GreenMask3,3,gn3)
                cG3=Transpose(Conjg(bG3))
		IF (LeastSquaring) THEN
!
		  CALL IncidentFields(Ei1,r0,Orientations,1)
		  CALL IncidentFields(Ei2,r0,Orientations,2)
!
! to circumvent a segment violation bug in DEC f90 1.0
!   matsqr3 faster than matmul(transp etc. on risc6000
!
!
 	     CALL matsqr3(A,bG3,gridvol3(ggg1,ggg2,ggg3,grid0))  
!
!
 		  e0(:,:,1)=e0(:,:,1)-
     &                matmul3(cG3,Ei1*gridvol3(ggg1,ggg2,ggg3,grid0))
		  e0(:,:,2)=e0(:,:,2)-
     &                matmul3(cG3,Ei2*gridvol3(ggg1,ggg2,ggg3,grid0))
		ENDIF
	      ELSE
		  bG3=0.0;cG3=0.0
	      ENDIF
!
	      bGV2=0.0
	      DO gg1=(ggg1-1)*gn2+1,ggg1*gn2
              DO gg2=(ggg2-1)*gn2+1,ggg2*gn2
              DO gg3=(ggg3-1)*gn2+1,ggg3*gn2
	       IF (gridMask2(gg1,gg2,gg3,grid0)) THEN
		r0=cell(c0)%center+posi2(gg1,gg2,gg3,:,grid0)
		GreenMask2=FarMask(r0,c0,2).AND..NOT.GreenMask3
		IF (Any(GreenMask2)) THEN
		  bG2=Greenint(dim,c0,r0,gg1,gg2,gg3,GreenMask2,2,gn2*gn3)
                  cG2=Transpose(Conjg(bG2))
		  bGV2=bGV2+bG2*gridvol2(gg1,gg2,gg3,grid0)
		  IF (LeastSquaring) THEN
!
		    CALL IncidentFields(Ei1,r0,Orientations,1)
		    CALL IncidentFields(Ei2,r0,Orientations,2)
!
! to circumvent a segment violation bug in DEC f90 1.0
!
! 		    IF (dim>=1000) THEN    
!
 		       CALL matsqr3(A,bG2,gridvol2(gg1,gg2,gg3,grid0))  
!
!		    ELSE
!		       A=A+matmul3(transpose(conjg(bG2)),bG2)*
!    &                   gridvol2(gg1,gg2,gg3,grid0) 
!		    ENDIF
!
		    e0(:,:,1)=e0(:,:,1)-matmul3(cG2,Ei1)*
     &                   gridvol2(gg1,gg2,gg3,grid0)
		    e0(:,:,2)=e0(:,:,2)-matmul3(cG2,Ei2)*
     &                   gridvol2(gg1,gg2,gg3,grid0)
		  ENDIF
		ELSE
		  bG2=0.0;cG2=0.0
		ENDIF
		bGV1=0.0
		DO g1=(gg1-1)*gn1+1,gg1*gn1
                DO g2=(gg2-1)*gn1+1,gg2*gn1
                DO g3=(gg3-1)*gn1+1,gg3*gn1 
		  IF (gridMask(g1,g2,g3,grid0)) THEN
		    GreenMask=(.NOT.GreenMask3).AND..NOT.GreenMask2
		    gvol=gridvol(g1,g2,g3,grid0)
		    vol=vol+gvol
		    r0=cell(c0)%center+posi(g1,g2,g3,:,grid0)
		    bG1=Greenint(dim,c0,r0,g1,g2,g3,GreenMask,1,gn1*gn2*gn3)
                    cG1=Transpose(Conjg(bG1))
		    bGV1=bGV1+bG1*gvol
		    basis=invsuski(g1,g2,g3,grid0)*
     &              baski(g1,g2,g3,:,1:cell(c0)%exporder,grid0)
     &              -matmul3(SingularValue3(:,:,g1,g2,g3,grid0),
     &                baski(g1,g2,g3,:,1:cell(c0)%exporder,grid0))
		    IF (LeastSquaring) THEN
			casis=TRANSPOSE(conjg(basis))
		    ELSE
			casis=Transpose(Conjg(invsuski(g1,g2,g3,grid0)*
     &                  baski(g1,g2,g3,:,1:cell(c0)%exporder,grid0)))
		    ENDIF
!
		    CALL IncidentFields(Ei1,r0,Orientations,1)
		    CALL IncidentFields(Ei2,r0,Orientations,2)
!
		    e1(:,:,1) = e1(:,:,1)+matmul3(casis,Ei1)
		    e1(:,:,2) = e1(:,:,2)+matmul3(casis,Ei2)
		    IF (LeastSquaring) THEN
		       e0(:,:,1) = e0(:,:,1)-matmul3(cG1,Ei1)*gvol
		       e0(:,:,2) = e0(:,:,2)-matmul3(cG1,Ei2)*gvol
!-------------------------------------------------------------------------
!		       IF (dim>=1000) THEN    
! to circumvent a segment violation bug in DEC f90 1.0
!
 			 CALL matsqr3(A,bG1,gvol) 
!
!		       ELSE
!			 A=A+matmul3(transpose(conjg(bG1)),bG1)*gvol 
!		       ENDIF           
!---------------------------------------------------------------------------
		    ENDIF
		    A1=A1-matmul3(casis,bG1+bG2+bG3)
		    A2=A2+matmul3(casis,basis)/gvol
		  ENDIF
		ENDDO;ENDDO;ENDDO
!        IF (Any(GreenMask2)) A(:,:)=A(:,:)+(matmul3(cG2,bGV1)+
!    &          conjg(transpose(matmul3(cG2,bGV1))))
		IF (Any(GreenMask2).AND.LeastSquaring)    
!
     &                   CALL matmul3cc(A,cG2,bGV1)
		bGV2=bGV2+bGV1
!
	       ENDIF
	      ENDDO;ENDDO;ENDDO
!       IF (Any(GreenMask3)) A(:,:)=A(:,:)+(matmul3(cG3,bGV2)+
!    &          conjg(transpose(matmul3(cG3,bGV2))))
	       IF (Any(GreenMask3).AND.LeastSquaring) 
!
     &                     CALL matmul3cc(A,cG3,bGV2)
!
	      ENDIF
	     ENDDO;ENDDO;ENDDO
!
! end of gridpoint-loop
!
	    WRITE(*,'(A,I4,F10.1)') '#',c0,CPUtime()
	    e0(cell(c0)%first:cell(c0)%last,:,1) =
     &          e0(cell(c0)%first:cell(c0)%last,:,1)+e1(:,:,1)
	    e0(cell(c0)%first:cell(c0)%last,:,2) =
     &          e0(cell(c0)%first:cell(c0)%last,:,2)+e1(:,:,2)
	    A(cell(c0)%first:cell(c0)%last,:)=
     &          A(cell(c0)%first:cell(c0)%last,:)+A1
	    IF (LeastSquaring) A(:,cell(c0)%first:cell(c0)%last)=
     &        A(:,cell(c0)%first:cell(c0)%last)+conjg(transpose(A1))
           A(cell(c0)%first:cell(c0)%last,cell(c0)%first:cell(c0)%last)=
     &     A(cell(c0)%first:cell(c0)%last,
     &     cell(c0)%first:cell(c0)%last)+A2
	    DEALLOCATE(casis,basis)
	    DEALLOCATE(A2)
	    DEALLOCATE(A1)
	    DEALLOCATE(e1)
	 ENDDO
      ENDIF
!
! end of cell-loop
!
	 write(*,*)vol
	 DEALLOCATE(SingularValue3)
	 END SUBROUTINE FillTheMatrices


!********************************************************************
	  FUNCTION Greenint(dim,c0,r0,sing1,sing2,sing3,greenMask,
     &      level,gnn)   
!********************************************************************
	  INTEGER,INTENT(in)::c0,dim,sing1,sing2,sing3,level,gnn
	  COMPLEX*16 Greenint(3,dim)
	  REAL*4,INTENT(in)::r0(3)
	  LOGICAL,INTENT(in),DIMENSION(0:cdim)::greenMask
	  LOGICAL,DIMENSION(cell(c0)%gridpointspcell,
     &     cell(c0)%gridpointspcell,cell(c0)%gridpointspcell)
     &     ::singMask
	  REAL*4,DIMENSION(gnn,gnn,gnn,3)::dr
	  INTEGER c,grid
	  REAL*4 vol

!  write(*,*)c0,level,cell(c0)%gridpoints,cell(c0)%exporder
!  write(*,*)GreenMask
       singMask=.TRUE.
       vol=0.0
!
       DO c=0,cdim
	 IF (GreenMask(c)) THEN
	    grid=cell(c)%grid
	    dr=0.0
	    SELECT CASE(level)
	    CASE(1)
	      WHERE (gridMask(:,:,:,grid))
		dr(:,:,:,1)=posi(:,:,:,1,grid)+cell(c)%center(1)-r0(1)
		dr(:,:,:,2)=posi(:,:,:,2,grid)+cell(c)%center(2)-r0(2)
		dr(:,:,:,3)=posi(:,:,:,3,grid)+cell(c)%center(3)-r0(3)
	      ENDWHERE
	      IF (c==c0) singMask(sing1,sing2,sing3)=.FALSE.     
	      Greenint(:,cell(c)%first:cell(c)%last)=
     &        Integrandi(cell(c)%exporder,gnn,baski(:,:,:,:,:,grid),
     &        dr,(gridMask(:,:,:,grid).AND.singMask),
     &        grs(:,:,:,:,grid),level)
	      IF (c==c0) singMask(sing1,sing2,sing3)=.TRUE.
	    CASE(2)
	      WHERE (gridMask2(:,:,:,grid))
		dr(:,:,:,1)=posi2(:,:,:,1,grid)+cell(c)%center(1)-r0(1)
		dr(:,:,:,2)=posi2(:,:,:,2,grid)+cell(c)%center(2)-r0(2)
		dr(:,:,:,3)=posi2(:,:,:,3,grid)+cell(c)%center(3)-r0(3)
	      ENDWHERE
	      Greenint(:,cell(c)%first:cell(c)%last)=
     &        Integrandi(cell(c)%exporder,gnn,baski2(:,:,:,:,:,grid),
     &         dr,gridMask2(:,:,:,grid),grs2(:,:,:,:,grid),level)
	    CASE(3)
	      WHERE (gridMask3(:,:,:,grid))
		dr(:,:,:,1)=posi3(:,:,:,1,grid)+cell(c)%center(1)-r0(1)
		dr(:,:,:,2)=posi3(:,:,:,2,grid)+cell(c)%center(2)-r0(2)
		dr(:,:,:,3)=posi3(:,:,:,3,grid)+cell(c)%center(3)-r0(3)
	      ENDWHERE
	      Greenint(:,cell(c)%first:cell(c)%last)=
     &          Integrandi(cell(c)%exporder,gnn,baski3(:,:,:,:,:,grid),
     &             dr,gridMask3(:,:,:,grid),grs3(:,:,:,:,grid),level)
	    CASE DEFAULT; Greenint=-Huge(DBLE(1.0))
	    END SELECT
	 ELSE
	      Greenint(:,cell(c)%first:cell(c)%last)= 0.0
	 ENDIF           
       ENDDO

	  END FUNCTION Greenint


!+++++++++++++++++++++++++++++++++++++++++++++++++++
	   FUNCTION FarMask(r0,c0,level)
!+++++++++++++++++++++++++++++++++++++++++++++++++++
	 LOGICAL,DIMENSION(0:cdim)::FarMask
	 INTEGER,INTENT(in)::c0,level
	 REAL*4,DIMENSION(3),INTENT(in)::r0
	 REAL*4 gridsize
	 REAL*4 rr
	 INTEGER c
	 
	   IF (level==1) THEN
	      FarMask=.TRUE.
	   ELSE 
	   DO c=0,cdim
	    IF (c==c0) THEN
	       FarMask(c)=.false.
	    ELSE
	       SELECT CASE(level)
		 CASE (1)
                  gridsize=maxval(gridvol (:,:,:,cell(c)%grid),
     &            Mask=gridMask (:,:,:,cell(c)%grid))**0.333333
		 CASE (2)
                  gridsize=maxval(gridvol2(:,:,:,cell(c)%grid),
     &            Mask=gridMask2(:,:,:,cell(c)%grid))**0.333333
		 CASE (3)
                  gridsize=maxval(gridvol3(:,:,:,cell(c)%grid),
     &            Mask=gridMask3(:,:,:,cell(c)%grid))**0.333333
	       END SELECT

	       rr=dot_product(r0-cell(c)%center,r0-cell(c)%center)
	       IF ( (24.0*reqaccuracy*(rr+1.0)*rr < 
     &        gridsize**2*(2.0*rr+12.0)).OR. (rr<5.0*cell(c)%size**2))
     &         THEN
		  FarMask(c)=.false.
	       ELSE
		  FarMask(c)=.true.
	       ENDIF
	     ENDIF
	  ENDDO
	  ENDIF
	 END FUNCTION FarMask
	 

!+++++++++++++++++++++++++++++++++++++++++++++++++++
	 FUNCTION Integrandi(bdim,sn,baski,dr,Mask,grs,level)
!+++++++++++++++++++++++++++++++++++++++++++++++++++
! VERY SPEED CRITICAL, OPTIMIZE IF YOU CAN!

! Optimal code depends a lot on your system architecture and 
! compiler capabilities.
! If you know what you are doing you may make significant speed
! improvements here.
! Find first how the temporaries are stored, and how the cache works
! and what common subexpressions are eliminated and are there any
! bank conflicts and so on.
	 

	  INTEGER,INTENT(in)::bdim,sn,level
	  COMPLEX*16,DIMENSION(sn,sn,sn,3,bdim),
     &     INTENT(in)::baski
	  LOGICAL,INTENT(in),DIMENSION(sn,sn,sn)::Mask
	  COMPLEX*16,DIMENSION(3,bdim)::Integrandi
	  COMPLEX*16,
     &     DIMENSION(sn,sn,sn):: Green1,Green2,summandi,RB
	  COMPLEX*16,DIMENSION(sn,sn,sn,3):: RG2
	  REAL*4,DIMENSION(sn,sn,sn,3),INTENT(in):: dR,grs
	  REAL*8,DIMENSION(sn,sn,sn,3):: Rhat
	  INTEGER i,j,k

	  LOGICAL,DIMENSION(sn,sn,sn,7)::NearMask
	  REAL*4,DIMENSION(sn,sn,sn)::R,RR
	  COMPLEX*16,DIMENSION(sn,sn,sn):: g
	  REAL*4,DIMENSION(sn,sn,sn,3)::ddr
	  REAL*4,DIMENSION(7,7)::gpt,gwt
	  INTEGER g1,g2,g3,j1,j2,j3,nd

	  gpt(1,1)=0.0; gwt(1,1)=2.0
	  gpt(1:2,2)=(/-0.5773502691, 0.5773502691/); gwt(1:2,2)=1.0
	  gpt(1:3,3)=(/-0.7745966692, 0.0,0.7745966692/)
          gwt(1:3,3)=(/0.555555555556,0.88888888889,0.55555555556/)
	  gpt(1:5,5)=(/-0.9061798459,-0.5348693101, 0.0,0.5348693101,
     &      0.9061798459/)
	  gwt(1:5,5)=(/ 0.2369268850, 0.4786286704, 0.568888888889,
     &      0.4786286704,0.2369268850/)
	  gpt(1:7,7)=(/-0.9491079123,-0.7415311855,-0.4058451513,
     &      0.0,0.4058451513,0.7415311855,0.9491079123/)
	  gwt(1:7,7)=(/ 0.1294849661, 0.2797053914, 0.3818300505,
     &      0.4179591836,0.3818300505,0.2797053914,0.1294849661/)


	  IF (sn==0) THEN
            do i=1,bdim
	    Integrandi(1,i)= 0.0
	    Integrandi(2,i)= 0.0
	    Integrandi(3,i)= 0.0
            enddo
	    STOP
	  ENDIF
	  NearMask=.false. 
!
             do i=1,sn ; do j=1,sn ;do k=1,sn
	  IF (Mask(k,j,i))THEN
	     RR(k,j,i)=dr(k,j,i,1)**2+dr(k,j,i,2)**2+dr(k,j,i,3)**2
             R(k,j,i)=sqrt(RR(k,j,i))
	     Rhat(k,j,i,1)=dr(k,j,i,1)/R(k,j,i)
	     Rhat(k,j,i,2)=dr(k,j,i,2)/R(k,j,i)
	     Rhat(k,j,i,3)=dr(k,j,i,3)/R(k,j,i)
	  ELSE
             Rhat(k,j,i,1)=0.0
             Rhat(k,j,i,2)=0.0
             Rhat(k,j,i,3)=0.0
             R(k,j,i)=0.0
	  ENDIF
             enddo; enddo; enddo
!
	  IF (level==1) THEN
	   WHERE (Mask)
	     NearMask(:,:,:,1)=
     &      (max(grs(:,:,:,1),max(grs(:,:,:,2),grs(:,:,:,3)))*2.5<R)
	     NearMask(:,:,:,3)=
     &      (max(grs(:,:,:,1),max(grs(:,:,:,2),grs(:,:,:,3)))*1.77<R)
     &      .AND..NOT. NearMask(:,:,:,1)
	     NearMask(:,:,:,5)=
     &      (max(grs(:,:,:,1),max(grs(:,:,:,2),grs(:,:,:,3)))*0.9<R) 
     &      .AND.(.NOT. NearMask(:,:,:,3))
     &      .AND.(.NOT.NearMask(:,:,:,1))
	     NearMask(:,:,:,7)=
     &       (.NOT. NearMask(:,:,:,5)).AND.(.NOT. NearMask(:,:,:,3))
     &       .AND.(.NOT.NearMask(:,:,:,1))
	   ENDWHERE 
	  ELSE
	   WHERE (Mask)
	     NearMask(:,:,:,1)=.TRUE.
	     NearMask(:,:,:,3)=.FALSE.
	     NearMask(:,:,:,5)=.FALSE.
	     NearMask(:,:,:,7)=.FALSE.
	   ENDWHERE 
	  ENDIF
!
	  WHERE (NearMask(:,:,:,1))
	     g=exp(im*R)/(RR*R*fourpi)   
	     Green1=(RR+im*R-DBLE(1.0))*g
	     Green2=(-RR-im*DBLE(3.0)*R+DBLE(3.0))*g
	     RG2(:,:,:,1)= Rhat(:,:,:,1)*Green2
	     RG2(:,:,:,2)= Rhat(:,:,:,2)*Green2
	     RG2(:,:,:,3)= Rhat(:,:,:,3)*Green2
	  ENDWHERE 
!
          DO j=1,bdim
	     WHERE (NearMask(:,:,:,1))
               RB=Rhat(:,:,:,1)*baski(:,:,:,1,j)+
     &            Rhat(:,:,:,2)*baski(:,:,:,2,j)+
     &            Rhat(:,:,:,3)*baski(:,:,:,3,j)
	     ENDWHERE 
             DO i=1,3
	       WHERE (NearMask(:,:,:,1))
		 summandi= baski(:,:,:,i,j)*Green1(:,:,:) +
     &                     RG2(:,:,:,i)*RB 
	       ENDWHERE 
               Integrandi(i,j)= sum(summandi,
     &                          Mask=NearMask(:,:,:,1)) 
	     ENDDO
          ENDDO

	  DO nd=3,7,2
	   IF (Any(NearMask(:,:,:,nd))) THEN
	    DO j1=1,nd;DO j2=1,nd;DO j3=1,nd
!
	     WHERE (NearMask(:,:,:,nd))
	       ddr(:,:,:,1)=dr(:,:,:,1)+gpt(j1,nd)*grs(:,:,:,1)*REAL(0.5)
	       ddr(:,:,:,2)=dr(:,:,:,2)+gpt(j2,nd)*grs(:,:,:,2)*REAL(0.5)
	       ddr(:,:,:,3)=dr(:,:,:,3)+gpt(j3,nd)*grs(:,:,:,3)*REAL(0.5)
	       RR=ddr(:,:,:,1)**2+ddr(:,:,:,2)**2+ddr(:,:,:,3)**2
!              DO i=1,sn;DO j=1,sn ;DO k=1,sn
!                 R(k,j,i)=sqrt(RR(k,j,i))
!              ENDDO; ENDDO; ENDDO
                  R=sqrt(RR)
	       g=exp(im*R)/(RR*R*fourpi)*gwt(j1,nd)*
     &           gwt(j2,nd)*gwt(j3,nd)/DBLE(8.0)
	       Green1=( RR+im*R-DBLE(1.0))*g
	       Green2=(-RR-im*DBLE(3.0)*R+DBLE(3.0))*g/RR
	       RG2(:,:,:,1)= ddr(:,:,:,1)*Green2
	       RG2(:,:,:,2)= ddr(:,:,:,2)*Green2
	       RG2(:,:,:,3)= ddr(:,:,:,3)*Green2
!
	     ENDWHERE 
	     DO j=1,bdim
!
	       WHERE (NearMask(:,:,:,nd))
                 RB=ddr(:,:,:,1)*baski(:,:,:,1,j)+
     &              ddr(:,:,:,2)*baski(:,:,:,2,j)+
     &              ddr(:,:,:,3)*baski(:,:,:,3,j)
!
	       ENDWHERE 
	       DO i=1,3
!
	         WHERE (NearMask(:,:,:,nd))
                   summandi= baski(:,:,:,i,j)*Green1(:,:,:) +
     &                       RG2(:,:,:,i)*RB 
	         ENDWHERE 
!
                 Integrandi(i,j)=  Integrandi(i,j) +
     &             sum(summandi,Mask=NearMask(:,:,:,nd)) 
	       ENDDO
              ENDDO
	    ENDDO;ENDDO;ENDDO
	   ENDIF
	 ENDDO

	 END FUNCTION Integrandi
!
!*****************************************************************
      SUBROUTINE DoSingularities3(cdim,gdim,gridpointspcell,
     &  SingularValue,grs)
!*****************************************************************

	 INTEGER,INTENT(in)::cdim,gdim,gridpointspcell
!
! cdim: cells, gdim: non-redundant cells
! gridpointspcell: grains-div/cell/direction
!
	 COMPLEX*16,INTENT(out),
     &    DIMENSION(3,3,gridpointspcell,gridpointspcell,
     &    gridpointspcell,1:gdim)::SingularValue
	 COMPLEX*16,DIMENSION(3,3):: singcorr1,singcorr2
	 LOGICAL done(gridpointspcell,gridpointspcell,
     &    gridpointspcell,0:cdim)
	 INTEGER c,g1,g2,g3,gg1,gg2,gg3,cc,i,j
	 REAL*4 vol
	 REAL*4,DIMENSION(:,:,:,:,:)::grs
!
!-----------------------------------------------------------------
!
! Note, requires modification for gradient index medium
	 SingularValue=0.0
	 done=.false.
!
! cell-loop
	  DO c=0,cdim
!
! only non-redundant cells
!
	    IF (.not.cell(c)%redundant) THEN
!
! gridpoint-loops
!
	      DO g3=1,cell(c)%gridpointspcell
              DO g2=1,cell(c)%gridpointspcell
              DO g1=1,cell(c)%gridpointspcell
		IF (gridMask(g1,g2,g3,cell(c)%grid)) THEN
!
		backloop: DO cc=0,c
!
                DO gg3=1,cell(cc)%gridpointspcell
                DO gg2=1,cell(cc)%gridpointspcell
                DO gg1=1,cell(cc)%gridpointspcell
		  IF (gridMask(gg1,gg2,gg3,cell(cc)%grid).AND.
     &                done(gg1,gg2,gg3,cc).AND.
     &                all(grs(g1,g2,g3,:,cell(c)%grid)==
     &                grs(gg1,gg2,gg3,:,cell(cc)%grid))) THEN
		    SingularValue(:,:,g1,g2,g3,cell(c)%grid)=
     &            SingularValue(:,:,gg1,gg2,gg3,cell(cc)%grid)
		    done(g1,g2,g3,c)=.true.
		    EXIT backloop
		  ENDIF
		 ENDDO;ENDDO;ENDDO
!
                ENDDO  backloop
!
		 IF (.NOT.done(g1,g2,g3,c)) THEN
!
		    CALL Singularity3(singcorr1,singcorr2,pointsize=
     &               grs(g1,g2,g3,:,cell(c)%grid))
!
! WHAT's WRONG BELOW? DEC compiles, but NAG stops.
!
		    SingularValue(:,:,g1,g2,g3,cell(c)%grid)=0.0
!
		    DO i=1,3
		      SingularValue(i,i,g1,g2,g3,cell(c)%grid)=
     &                 -DBLE(1.0)/DBLE(3.0)+singcorr1(i,i)-
     &                 cell(c)%meanindex**2*singcorr2(i,i)
		    ENDDO
!
!    write(*,*)'singularity',c,cell(c)%grid,SingularValue(cell(c)%grid,sisi),
!            SingularValue(cell(c)%grid,sisi)*vol
!
		    done(g1,g2,g3,c)=.true.
		 ENDIF
		ENDIF
	      ENDDO;ENDDO;ENDDO
!
! end gridpoint-loops
!
	    ENDIF
!
	  ENDDO
!
! end cell-loop
!
	 END SUBROUTINE DoSingularities3

!***************************************************************************
	SUBROUTINE Singularity3(singcorr1,singcorr2,pointsize) !NOT YET OK
!***************************************************************************
	COMPLEX*16,INTENT(out),
     &   DIMENSION(3,3):: singcorr1,singcorr2  !diag tensor
	REAL*4,INTENT(in):: pointsize(3)
	REAL*4 RR,innradius,bb
	REAL*4,DIMENSION(3):: R,Rhat
	INTEGER k1,k2,k3,i,j
	INTEGER,PARAMETER::k=33
	COMPLEX*16 Green1,Green2

	 bb=(maxval(pointsize)/k)**3
	 innradius=minval(pointsize)*0.5
	 singcorr1=0.0
	 singcorr2=0.0

      DO k1=1,k; DO k2=1,k; DO k3=1,k
       R=maxval(pointsize)*(-0.5+((/k1,k2,k3/)-0.5)/k)
       IF (all(R<0.5*pointsize).AND.all(R>-0.5*pointsize)) THEN
         RR=dot_product(R,R)
         IF (RR > innradius**2) THEN
 
          CALL Greeni(R,Rhat,Green1,Green2) 
 
          DO i=1,3 
           singcorr1(i,i)=singcorr1(i,i)+(Green1+Rhat(i)**2 *Green2)*bb
             IF (all(R<0.5*minval(pointsize)).AND.
     &            all(R>-0.5*minval(pointsize))) THEN    
 	       singcorr2(i,i)=singcorr2(i,i)  
     &         +Green1*R(i)**2*bb*0.5 
     &         +Green2*Rhat(i)**2*(2.0*R(i)**2-sum(R**2))*bb*0.5
!             +Green1*R(2)**2*bb*0.5 &
!             +Green2*Rhat(2)**2*(R(2)**2-2.0*R(3)**2)*bb*0.5
             ELSE
! Yes, there really should be something here, but it is too complicated for me.
! Hope nobody will notice ...
             ENDIF 
          ENDDO
        ENDIF
       ENDIF
          ENDDO;ENDDO;ENDDO
!
	 DO i=1,3;j=i
          singcorr1(i,j)=singcorr1(i,j)+
     &   2.0/3.0*((DBLE(1.0)-im*innradius)*exp(im*innradius)-DBLE(1.0))
	  singcorr2(i,j)=singcorr2(i,j)+(DBLE(1.0)-exp(im*innradius))
     &     +exp(im*innradius)*(-2.0/15.0*im*innradius**2+
     &      7.0/15.0*innradius+im)*innradius
	 ENDDO

	END SUBROUTINE Singularity3


!******************************************************************
	  SUBROUTINE Greeni(dr,Rhat,Green1,Green2)
!******************************************************************
	  REAL*4,DIMENSION(3),INTENT(in)::dr
	  REAL*4,DIMENSION(3),INTENT(out)::Rhat
	  COMPLEX*16,INTENT(out)::Green1,Green2
	  REAL*4 R,RR
	  COMPLEX*16 g
	     RR=dot_product(dr,dr);R=sqrt(RR); Rhat=dr/R
	     g=exp(im*R)/(RR*R*fourpi)   
	     Green1=( RR+im*R-DBLE(1.0))*g 
	     Green2=(-RR-im*DBLE(3.0)*R+DBLE(3.0))*g 
	  END SUBROUTINE Greeni

       END MODULE FillMatrices

!===========================================================================
!===========================================================================
