      MODULE Incidentia
! Only to give the incident Field vectors
!
       USE Constants
       USE Vectorix,ONLY:OPERATOR(.CROSS.),NORMALIZE
!
       IMPLICIT NONE

       REAL*4,SAVE,allocatable,dimension(:,:)::k0,ip1,ip2  
! wavevector and incident polarization vectors
!
       REAL*4,SAVE,public::wavelength=1.0 ,wavenumber=twopi


      CONTAINS

!*****************************************************************
!
       SUBROUTINE GetWavelength()
           WRITE (*,*) 'Wavelength?  '
           WRITE (*,*) '(0 =>2pi is not a bad choice, '
           WRITE (*,*) ' unless you know better)'
           READ(15,*) Wavelength
           IF (wavelength<=REAL(0.0)) wavelength=twopi
           wavenumber=twopi/wavelength
           WRITE(*,*) Wavelength, wavenumber
         END SUBROUTINE Getwavelength
 
!*****************************************************************
!
       SUBROUTINE Incidences(Orientations,OriAveType)
! to create the fields     
!
           INTEGER,intent(in)::Orientations,OriAveType
           INTEGER o
           REAL*4,DIMENSION(3)::k00,ip0,b1,b2
 
           ALLOCATE(k0(3,Orientations),ip1(3,Orientations),
     &      ip2(3,Orientations))
!
           IF (oriAveType==0) THEN
            DO o=1,Orientations
             WRITE (*,*) 'The wave vector of the incident field '
             WRITE (*,*) '(the opposite of the direction of source)?'
             READ (15,*) k00(1),k00(2),k00(3)
             WRITE(*,*) k00(1),k00(2),k00(3)
             WRITE (*,*) 'Just another vector in the scattering plane?'
             READ (15,*) ip0(1),ip0(2),ip0(3)
             WRITE (*,*) ip0(1),ip0(2),ip0(3)
             k0(:,o)=Normalize(k00)
             ip2(:,o)=Normalize(ip0.CROSS.k0(:,o))
             ip1(:,o)=Normalize(ip2(:,o).CROSS.k0(:,o))
            END DO
!
           ELSE IF (oriAveType==1) THEN
            DO o=1,Orientations
             DO
               CALL Random_Number(k0(:,o))
               IF (dot_product(k0(:,o),k0(:,o)) <= REAL(1.0)) EXIT
             ENDDO 
             DO
               CALL Random_Number(ip0)
               IF (dot_product(ip0,ip0) <= REAL(1.0)) EXIT
             ENDDO 
             ip1(:,o)=Normalize(k0(:,o).CROSS.ip0)
             ip2(:,o)=Normalize(ip1(:,o).CROSS.k0(:,o))
             k0(:,o)=Normalize(ip2(:,o).CROSS.ip1(:,o))  !*2*pi/wavelength
            END DO
!
           ELSE IF (oriAveType==2) THEN
            WRITE(*,*) 'Please, define the rotation axis.'
            READ (*,*) ip0(1),ip0(2),ip0(3)
            b1=Normalize(ip0.cross.(/REAL(0.1239282),REAL(0.73998739),
     &        -REAL(0.82926323761821)/))
            b2=Normalize(ip0.cross.b1)
!
            DO o=1,Orientations
              ip1(:,o)=ip0
              ip2(:,o)=Normalize(cos((o-1)*pi*2/(Orientations-1.0))*b1
     &        +sin((o-1)*pi*2/(Orientations-1.0))*b2)
              k0(:,o)=Normalize(ip2(:,o).CROSS.ip1(:,o))  
            END DO
!
           ENDIF
!
        END SUBROUTINE Incidences


!*********************************************************************
!
        SUBROUTINE IncidentFields(Einc,r,oriens,pp)
         INTEGER,INTENT(in)::oriens,pp
         COMPLEX*16,DIMENSION(3,oriens):: Einc
         REAL*4,DIMENSION(3),INTENT(in)::r
         INTEGER o

        DO o=1,oriens
        two_polarizations:SELECT CASE(pp)  
          CASE (1)
           EInc(:,o)=exp(im*dot_product(k0(:,o),r))*ip1(:,o)
          CASE (2)
           EInc(:,o)=exp(im*dot_product(k0(:,o),r))*ip2(:,o)
          CASE DEFAULT
           STOP 'This cannot happen, at least in IncidentFields'
        END SELECT two_polarizations
        ENDDO
!
! By the way, if you have another kind of incident field as plane wave,
! e.g. a beam, go ahead and dumb it here. It works!
!
        END SUBROUTINE IncidentFields

!*******************************************************************
!
        SUBROUTINE IncidentFields2(Einc,r,gridmask,gn,oriens,pp)
!
         INTEGER,INTENT(in)::oriens,pp,gn
         COMPLEX*16,DIMENSION(gn,gn,gn,3,oriens):: Einc
         REAL*4,DIMENSION(gn,gn,gn,3),INTENT(in)::r
         LOGICAL,DIMENSION(gn,gn,gn),INTENT(in)::gridmask
         INTEGER o,g1,g2,g3

        DO o=1,oriens
        DO g3=1,gn; DO g2=1,gn; DO g1=1,gn
        IF (gridmask(g1,g2,g3)) THEN
        two_polarizations:SELECT CASE(pp)  
          CASE (1)
            EInc(g1,g2,g3,:,o)=exp(im*dot_product(k0(:,o),
     &                            r(g1,g2,g3,:)))*ip1(:,o)
          CASE (2)
            EInc(g1,g2,g3,:,o)=exp(im*dot_product(k0(:,o),
     &                            r(g1,g2,g3,:)))*ip2(:,o)
          CASE DEFAULT
           STOP 'This cannot happen, at least in IncidentFields'
        END SELECT two_polarizations
        ELSE    
          Einc(g1,g2,g3,:,o)=0.0
        ENDIF
        ENDDO;ENDDO;ENDDO
        ENDDO
! By the way, if you have another kind of incident field as plane wave,
! e.g. a beam, go ahead and dumb it here. It works!
        END SUBROUTINE IncidentFields2

      END MODULE Incidentia
