! **************************************************************************
	MODULE Resultia
! *************************************************************************

	 USE Constants
	 USE Incidentia,
     &    ONLY:k0,ip1,ip2,wavenumber,wavelength,IncidentFields2
	 USE Scatterer,ONLY:subgrains,grainclass
	 USE DEfBaseGrid,
     &   ONLY:cdim,cell,posi,gridmask,gridvol,baski,invsuski,reqaccuracy
	 USE Timing,ONLY:CPUtime

	 IMPLICIT NONE

! PUBLIC::   StoreResults,FinalOutput,InitDetectors





	 INTEGER,PARAMETER::angles=37   
	 REAL*4,
     &    DIMENSION(angles,4,4)::ScatteringMatrix=0.0,ScatMat2=0.0
	 REAL*4,SAVE,
     &    DIMENSION(angles)::ScatteringAngle,PhaseAngle,AngleWeight
	 REAL*4:: Cext=0.0,Csca=0.0,Cabs=0.0,albedo=0.0,
     &    Cpha=0.0,Cpol=0.0,Cbak=0.0
	 REAL*4:: Cext2=0.0,Cabs2=0.0,Cpol2=0.0,
     &    Csca2=0.0,Cbak2=0.0
	 REAL*4,SAVE,DIMENSION(:,:,:),
     &    ALLOCATABLE::Obs,Ep1,Ep2
!
! Observation direction and polarization vectors
! INTEGER:: outfile=11
! CHARACTER(len=33) ::outfilename  ='jscat.results'

	 REAL*4,SAVE::cvol=0.0
	 INTEGER,PARAMETER::thetafix=9
	 REAL*4,SAVE,DIMENSION(111,2)::LightCurve

	 CONTAINS

!  **************************************************************************
	  SUBROUTINE StoreResults(config,ncaver,dim,Oriens,polaris,e,
     &               outfile)
!  **************************************************************************
! called after finishing with each configuration

	    INTEGER, INTENT(IN)::Oriens,config,ncaver,dim,polaris,
     &     outfile
	    COMPLEX*16,INTENT(IN),DIMENSION(dim,Oriens,polaris)::e

	    complex*16,DIMENSION(angles,4,Oriens)::AmpScatMat  
! AmplitudeScatteringMatrix
	    REAL*4,DIMENSION(angles,4,4)::ScatteringMat1
	    REAL*4,DIMENSION(ncaver,Oriens)::pdf 
!
! propability density function for configurations
! pdf=pdf(conf,o), pdf >= 0.0, sum_conf,o=1^ncaver(pdf)=1.0
!
	    REAL*4,DIMENSION(Oriens):: CCabs,CCext,CCpol,CCbak
	    INTEGER ang,ori
	    
	  ALLOCATE(obs(3,angles,Oriens),Ep1(3,angles,Oriens),
     &     Ep2(3,angles,Oriens))
!
	  CALL GetObservingGeometry(angles,Oriens,Ep1,Ep2,obs)
!
	  pdf=REAL(1.0)/(ncaver*Oriens) ! find better
!
	  CALL ComputeCrossSections(CCext,CCabs,CCpol,CCbak,
     &   AmpScatMat,cvol,obs,Ep1,Ep2,angles,dim,Oriens,polaris,e)

	  Orientations: DO ori=1, Oriens
	   Cext=Cext+CCext(ori)*pdf(config,ori)/wavenumber**2
	   Cabs=Cabs+CCabs(ori)*pdf(config,ori)/wavenumber**2
	   Cpol=Cpol+CCpol(ori)*pdf(config,ori)/wavenumber**2
	   Cbak=Cbak+CCbak(ori)*pdf(config,ori)/wavenumber**2
	   Cext2=Cext2+CCext(ori)**2*pdf(config,ori)/wavenumber**4
	   Cabs2=Cabs2+CCabs(ori)**2*pdf(config,ori)/wavenumber**4
	   Cpol2=Cpol2+CCpol(ori)**2*pdf(config,ori)/wavenumber**4
	   Cbak2=Cbak2+CCbak(ori)**2*pdf(config,ori)/wavenumber**4
! Until this point, unit system k==1 was assumed, but for output, results are scaled to normal units
	   DO ang=1,angles
	    ScatteringMat1(ang,1,1)=
     &      0.5*SNGL(DBLE(conjg(AmpScatMat(ang,1,ori))*
     &       AmpScatMat(ang,1,ori)
     &      +conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,2,ori)
     &      +conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,3,ori)
     &      +conjg(AmpScatMat(ang,4,ori))*AmpScatMat(ang,4,ori)))
            ScatteringMat1(ang,1,2)=
     &      0.5*SNGL(DBLE(-conjg(AmpScatMat(ang,1,ori))*
     &       AmpScatMat(ang,1,ori)
     &      +conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,2,ori)
     &      -conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,3,ori)
     &      +conjg(AmpScatMat(ang,4,ori))*AmpScatMat(ang,4,ori)))
             ScatteringMat1(ang,1,3)=
     &      REAL( conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,3,ori)
     &      +conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,4,ori))
	    ScatteringMat1(ang,1,4)=
     &      + AIMAG(conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,3,ori)
     &      -conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,4,ori))

	    ScatteringMat1(ang,2,1)=
     &      0.5*REAL(-conjg(AmpScatMat(ang,1,ori))
     &       *AmpScatMat(ang,1,ori)
     &       +conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,2,ori)
     &      +conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,3,ori)
     &      -conjg(AmpScatMat(ang,4,ori))*AmpScatMat(ang,4,ori))
	    ScatteringMat1(ang,2,2)=
     &      0.5*REAL(+conjg(AmpScatMat(ang,1,ori))
     &       *AmpScatMat(ang,1,ori)
     &      +conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,2,ori)
     &      -conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,3,ori)
     &      -conjg(AmpScatMat(ang,4,ori))*AmpScatMat(ang,4,ori))
	    ScatteringMat1(ang,2,3)=
     &      REAL( conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,3,ori)
     &      -conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,4,ori))
	    ScatteringMat1(ang,2,4)=
     &      + AIMAG(conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,3,ori)
     &      +conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,4,ori))

            ScatteringMat1(ang,3,1)=
     &     REAL( conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,4,ori)
     &     +conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,3,ori))
	    ScatteringMat1(ang,3,2)=
     &     + REAL( conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,4,ori)
     &      -conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,3,ori))
	    ScatteringMat1(ang,3,3)=
     &     + REAL( conjg(AmpScatMat(ang,2,ori))*AmpScatMat(ang,1,ori)
     &     +conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,4,ori))
	    ScatteringMat1(ang,3,4)=
     &       + AIMAG(-conjg(AmpScatMat(ang,2,ori))
     &       *AmpScatMat(ang,1,ori)
     &      -conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,4,ori))

           ScatteringMat1(ang,4,1)=
     &     AIMAG( conjg(AmpScatMat(ang,4,ori))*AmpScatMat(ang,2,ori)
     &      +conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,3,ori))
	    ScatteringMat1(ang,4,2)=
     &     +AIMAG( conjg(AmpScatMat(ang,4,ori))*AmpScatMat(ang,2,ori)
     &      -conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,3,ori))
	    ScatteringMat1(ang,4,3)=
     &     +AIMAG( conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,2,ori)
     &       -conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,4,ori))
	    ScatteringMat1(ang,4,4)=
     &     REAL( conjg(AmpScatMat(ang,1,ori))*AmpScatMat(ang,2,ori)
     &      -conjg(AmpScatMat(ang,3,ori))*AmpScatMat(ang,4,ori))
!
!
	    ENDDO
!
	    ScatteringMatrix=ScatteringMatrix+
     &       ScatteringMat1*pdf(config,ori)/wavenumber**2
	    ScatMat2=ScatMat2+ScatteringMat1**2
     &      *pdf(config,ori)/wavenumber**4
	    Csca=Csca+sum(ScatteringMat1(:,1,1)*
     &       AngleWeight(:))*twopi*pdf(config,ori)/wavenumber**2
	    Csca2=Csca2+(sum(ScatteringMat1(:,1,1)*
     &       AngleWeight(:))*twopi/wavenumber**2)**2*pdf(config,ori)
	    IF (ori<112) THEN
         LightCurve(ori,1)= 0.5*REAL(conjg(AmpScatMat(thetafix,1,ori))*
     &         AmpScatMat(thetafix,1,ori)
     &    +conjg(AmpScatMat(thetafix,3,ori))
     &       *AmpScatMat(thetafix,3,ori))  
         LightCurve(ori,2)= 0.5*REAL(conjg(AmpScatMat(thetafix,2,ori))*
     &         AmpScatMat(thetafix,2,ori)
     &    +conjg(AmpScatMat(thetafix,4,ori))*AmpScatMat(thetafix,4,ori))
	    ENDIF
	  ENDDO Orientations
	DEALLOCATE(obs,Ep1,Ep2)
	END SUBROUTINE StoreResults     

	 
! **************************************************************************
	  SUBROUTINE GetObservingGeometry(angles,Oriens,Ep1,Ep2,Obs)
! **************************************************************************
! to select the observing angles
!
	 INTEGER,INTENT(IN)::angles,Oriens
	 REAL*4,INTENT(out),
     &     DIMENSION(3,angles,Oriens)::Obs,Ep1,Ep2
! Observation direction and polarization vectors
	 INTEGER angl,o
	 REAL*4 sinTheta,cosTheta

	 DO angl=1,angles
!
	   ScatteringAngle(angl)=180.0/(angles-1.0)*(angl-1.0)
	   cosTheta=cos(ScatteringAngle(angl)*degree)
	   sinTheta=sin(ScatteringAngle(angl)*degree)
	   PhaseAngle(angl)=180.0-ScatteringAngle(angl)
	   AngleWeight(angl)=pi*sinTheta/(angles-1.0)
!
	   DO o=1,Oriens
	     Obs(:,angl,o)=cosTheta*k0(:,o)+sinTheta*ip1(:,o)
	     Ep1(:,angl,o)=(-sinTheta)*k0(:,o)+cosTheta*ip1(:,o)
	     Ep2(:,angl,o)=ip2(:,o)
!         write(*,*)'ori= ',o,' angle= ',angl
!         write(*,*)'obs= ',Obs(:,angl,o)
!         write(*,*)'Ep1= ',Ep1(:,angl,o)
!         write(*,*)'Ep2= ',Ep2(:,angl,o)
	   END DO 
!
	 END DO 
!
	 AngleWeight(1)=1.0-sum(AngleWeight(2:angles-1))*0.5
	 AngleWeight(angles)=2.0-sum(AngleWeight(1:angles-1))
	 AngleWeight=2.0*Angleweight/sum(AngleWeight)
! write(*,*) angleweight
!  write(*,*) '|Obs|',abs(Obs(angl))
!  write(*,*) 'Obs.dot.k0',Obs(angl).dot.k0(o),-cos(ScatteringAngle(angl)*degree)
!
	END SUBROUTINE GetObservingGeometry

!***************************************************************************
      SUBROUTINE ComputeCrossSections(Cext,Cabs,Cpol,Cbak,ASM,
     &    totvol,obs,Ep1,Ep2,angles,dim,Oriens,polaris,e)
!***************************************************************************

	INTEGER,INTENT(IN)::angles,dim,Oriens,polaris
	COMPLEX*16,INTENT(IN),
     &   DIMENSION(dim,Oriens,polaris)::e
! e=A*b solution!
!
	REAL*4,INTENT(IN),
     &   DIMENSION(3,angles,Oriens)::Obs,Ep1,Ep2
! Observation direction and polarization vectors
!
	COMPLEX*16,INTENT(out),
     &   DIMENSION(angles,4,Oriens)::ASM
! Amplitude scattering Matrix
!
	REAL*4,INTENT(out),
     &   DIMENSION(Oriens)::cext,cabs,cpol,cbak
	REAL*4,INTENT(out):: totvol
	COMPLEX*16,DIMENSION(:,:,:,:,:),
     &     ALLOCATABLE:: Einc1,Einc2,Pint1,Pint2
	REAL*4,DIMENSION(:,:,:,:),ALLOCATABLE::r
	COMPLEX*8,ALLOCATABLE:: fact(:,:,:)
	INTEGER o,c,grid,a
        INTEGER xx0 !dummy
        xx0=cell(c)%gridpointspCell/2

	Cabs=0.0;Cext=0.0;Cpol=0.0;Cbak=0.0;ASM=0.0;totvol=0.0
!
! subgrain-loop
!
	 DO c=0,cdim 
!
	  grid=cell(c)%grid
	  totvol=totvol+cell(c)%volume
!
          ALLOCATE(fact(cell(c)%gridpointspCell,cell(c)%gridpointspCell,
     &    cell(c)%gridpointspCell))
	  ALLOCATE(r(cell(c)%gridpointspCell,cell(c)%gridpointspCell,
     &    cell(c)%gridpointspCell,3))
!
	  r(:,:,:,1)=posi(:,:,:,1,grid)+cell(c)%center(1)
	  r(:,:,:,2)=posi(:,:,:,2,grid)+cell(c)%center(2)
	  r(:,:,:,3)=posi(:,:,:,3,grid)+cell(c)%center(3)
!
! allocate internal/incident fields
!
	  ALLOCATE(Pint1(cell(c)%gridpointspCell,
     &    cell(c)%gridpointspCell,cell(c)%gridpointspCell,3,
     &    Oriens))
          ALLOCATE(Pint2(cell(c)%gridpointspCell,
     &    cell(c)%gridpointspCell,cell(c)%gridpointspCell,3,
     &    Oriens))
	  ALLOCATE(Einc1(cell(c)%gridpointspCell,
     &    cell(c)%gridpointspCell,cell(c)%gridpointspCell,3,
     &         Oriens))
          ALLOCATE(Einc2(cell(c)%gridpointspCell,
     &    cell(c)%gridpointspCell,cell(c)%gridpointspCell,3,
     &         Oriens))
!
	     CALL InternalFields(Pint1,e,dim,Oriens,c,1)
	     CALL InternalFields(Pint2,e,dim,Oriens,c,2)
	     CALL IncidentFields2(Einc1,r,gridmask(:,:,:,grid),
     &                    cell(c)%gridpointspCell,Oriens,1)
	     CALL IncidentFields2(Einc2,r,gridmask(:,:,:,grid),
     &                    cell(c)%gridpointspCell,Oriens,2)
!
!     write(*,*) 'Pint1',Pint1(1,1)
!     write(*,*) 'Einc1',Einc1(1,1)     
!     write(*,*) 'Pint2',Pint2     
!     write( *) 'Einc2',Einc2     
!
! orientation-loop
!
	     DO o=1,Oriens
!
      cabs(o)=cabs(o)-0.5*sum(DBLE(conjg(PInt1(:,:,:,1,o))*
     &    Pint1(:,:,:,1,o)+conjg(Pint1(:,:,:,2,o))*Pint1(:,:,:,2,o)
     &         +conjg(PInt1(:,:,:,3,o))*Pint1(:,:,:,3,o)
     &         +conjg(PInt2(:,:,:,1,o))*Pint2(:,:,:,1,o)
     &         +conjg(PInt2(:,:,:,2,o))*Pint2(:,:,:,2,o)
     &         +conjg(PInt2(:,:,:,3,o))*Pint2(:,:,:,3,o))
     &         *gridvol(:,:,:,grid)*Aimag(invsuski(:,:,:,grid)),
     &                                 MASK=gridmask(:,:,:,grid))
      cext(o)=cext(o)+0.5*sum(Aimag(conjg(Einc1(:,:,:,1,o))
     &               *Pint1(:,:,:,1,o)
     &        +conjg(Einc1(:,:,:,2,o))*Pint1(:,:,:,2,o)
     &        +conjg(Einc1(:,:,:,3,o))*Pint1(:,:,:,3,o)
     &        +conjg(Einc2(:,:,:,1,o))*Pint2(:,:,:,1,o)
     &        +conjg(Einc2(:,:,:,2,o))*Pint2(:,:,:,2,o)
     &        +conjg(Einc2(:,:,:,3,o))*Pint2(:,:,:,3,o)
     &       )*gridvol(:,:,:,grid),MASK=gridmask(:,:,:,grid))
              cpol(o)=cpol(o)+0.5*
     &         Aimag(sum((conjg(Einc1(:,:,:,1,o))*Pint1(:,:,:,1,o)
     &         +conjg(Einc1(:,:,:,2,o))*Pint1(:,:,:,2,o)
     &         +conjg(Einc1(:,:,:,3,o))*Pint1(:,:,:,3,o)
     &         -conjg(Einc2(:,:,:,1,o))*Pint2(:,:,:,1,o)
     &         -conjg(Einc2(:,:,:,2,o))*Pint2(:,:,:,2,o)
     &         -conjg(Einc2(:,:,:,3,o))*Pint2(:,:,:,3,o)
     &          )*gridvol(:,:,:,grid),MASK=gridmask(:,:,:,grid)))
! just a test
	      cbak(o)=cbak(o)+0.5*
     &        Aimag(sum((Einc1(:,:,:,1,o)*Pint1(:,:,:,1,o)  
     &          +Einc1(:,:,:,2,o)*Pint1(:,:,:,2,o)  
     &          +Einc1(:,:,:,3,o)*Pint1(:,:,:,3,o)  
     &          +Einc2(:,:,:,1,o)*Pint2(:,:,:,1,o)  
     &          +Einc2(:,:,:,2,o)*Pint2(:,:,:,2,o)  
     &          +Einc2(:,:,:,3,o)*Pint2(:,:,:,3,o)  
     &           )*gridvol(:,:,:,grid),MASK=gridmask(:,:,:,grid)))
!
! scattering angle-loop
!
	      DO a=1,angles
		WHERE (gridmask(:,:,:,grid))
                 fact=exp(-im*(Obs(1,a,o)*r(:,:,:,1)+
     &             Obs(2,a,o)*r(:,:,:,2)+Obs(3,a,o)*
     &             r(:,:,:,3)))*(im/fourpi)*gridvol(:,:,:,grid)
		ELSEWHERE
                 fact=0.0
		ENDWHERE
	        ASM(a,1,o)=ASM(a,1,o)+sum(
     &                      fact*(Ep2(1,a,o)*PInt2(:,:,:,1,o)
     &                     +Ep2(2,a,o)*PInt2(:,:,:,2,o)
     &                     +Ep2(3,a,o)*PInt2(:,:,:,3,o)),
     &                      MASK=gridmask(:,:,:,grid))
	        ASM(a,2,o)=ASM(a,2,o)+sum(
     &                      fact*(Ep1(1,a,o)*PInt1(:,:,:,1,o)
     &                     +Ep1(2,a,o)*PInt1(:,:,:,2,o)
     &                     +Ep1(3,a,o)*PInt1(:,:,:,3,o)),
     &                      MASK=gridmask(:,:,:,grid))
	        ASM(a,3,o)=ASM(a,3,o)+sum(
     &                      fact*(Ep1(1,a,o)*PInt2(:,:,:,1,o)
     &                     +Ep1(2,a,o)*PInt2(:,:,:,2,o)
     &                     +Ep1(3,a,o)*PInt2(:,:,:,3,o)),
     &                      MASK=gridmask(:,:,:,grid))
	        ASM(a,4,o)=ASM(a,4,o)+sum(
     &                      fact*(Ep2(1,a,o)*PInt1(:,:,:,1,o)
     &                     +Ep2(2,a,o)*PInt1(:,:,:,2,o)
     &                     +Ep2(3,a,o)*PInt1(:,:,:,3,o)),
     &                      MASK=gridmask(:,:,:,grid))
!
!       write(*,*)'angle=',a,' ori= ',o
!       write(*,*)ASM(a,1,o)
!       write(*,*)'Ep2 =',Ep2(:,a,o)
!       write(*,*)'PInt2 =',PInt2(xx0,xx0,xx0,:,o)
!       write(*,*)ASM(a,2,o)
!       write(*,*)'Ep1 =',Ep1(:,a,o)
!       write(*,*)'PInt1 =',PInt1(xx0,xx0,xx0,:,o)
	      ENDDO ! end angle's loop
!
	     ENDDO !  end of orientational loop
!
	     DEALLOCATE(Einc2)
	     DEALLOCATE(Einc1)
	     DEALLOCATE(Pint2)
	     DEALLOCATE(Pint1)
	     DEALLOCATE(r)
	     DEALLOCATE(fact)
!
! end subgrain-loop
!
	  ENDDO 

	END SUBROUTINE ComputeCrossSections

! ***********************************************************
	  SUBROUTINE InternalFields(Pint,e,dim,Oriens,c,polaris)
! ***********************************************************
! Polarization field inside cell c at grid point g1,g2,g3 for all orientations
      INTEGER,INTENT(IN)::dim,Oriens,polaris,c
      COMPLEX*16,INTENT(IN),DIMENSION(dim,Oriens,polaris)::e
      COMPLEX*16,
     & DIMENSION(cell(c)%gridpointspCell,cell(c)%gridpointspCell,
     & cell(c)%gridpointspCell,3,Oriens),
     & INTENT(out):: Pint
      INTEGER g1,g2,g3

       DO g3=1,cell(c)%gridpointspCell
       DO g2=1,cell(c)%gridpointspCell
       DO g1=1,cell(c)%gridpointspCell
	IF (gridmask(g1,g2,g3,cell(c)%grid)) THEN
	 Pint(g1,g2,g3,:,:)=
     &   matmul(Baski(g1,g2,g3,:,1:cell(c)%exporder,cell(c)%grid),
     &    e(cell(c)%first:cell(c)%last,:,polaris))
     &    /gridvol(g1,g2,g3,cell(c)%grid ) 
	ELSE
	  Pint(g1,g2,g3,:,:)=0.0
	ENDIF
	  ENDDO;  ENDDO;  ENDDO       


	  END SUBROUTINE InternalFields


!   **************************************************************************
      SUBROUTINE FinalOutput(wavenumber,outfile)
!   **************************************************************************
      USE Mie
      INTEGER, INTENT(IN)::wavenumber
      INTEGER i,outfile
      REAL*4 asymmetry,apualbedo
      REAL*8 mieradius,mieradius1
      REAL*4,DIMENSION(angles,4,4)::PhaseMatrix,PhaseMatErr
      INTEGER timevalue(8)

      write (*,*)
      write (*,*) '==================================================='
      write (*,*)
      write (*,*) 'RESULTS: '
      write (*,*)
      PhaseMatrix=ScatteringMatrix*(fourpi/Csca)
      PhaseMatErr=sqrt(max(ScatMat2
     & -ScatteringMatrix**2,
     & REAL(0.0))/wavenumber)*(fourpi/Csca)
      asymmetry=sum(PhaseMatrix(:,1,1)*AngleWeight(:)
     & *cos(ScatteringAngle(:)*degree))*0.5
      albedo=Csca/Cext;   apualbedo=(Cext-Cabs)/Cext
!     OPEN (outfile,file=outfilename,form='unformatted',
!    &   status='unknown')
!
!     OPEN (outfile,file=outfilename,status='unknown')
!     write (outfile,*) angles
!     write (outfile,*) Cext,Cabs,Csca,albedo,Cpha,Cpol
!     write (outfile,*) ScatteringAngle,AngleWeight,PhaseMatrix 
!
      mieradius=(3*cvol/fourpi)**0.3333333333333333
!
      CALL BhMie(mieradius,grainclass(1)%refindx*DBLE(1.0),angles,
     & s11Mie,s12Mie,CextMie,CscaMie)
!
      s12Mie=s12Mie/sum(s11Mie(1:angles)*AngleWeight(:))*2
      s11Mie=s11Mie/sum(s11Mie(1:angles)*AngleWeight(:))*2
      CextMie=CextMie*pi*(mieradius/wavenumber)**2
      CscaMie=CscaMie*pi*(mieradius/wavenumber)**2
!
      mieradius1=(3*cvol/fourpi/subgrains)**0.3333333333333333
!
      CALL BhMie(mieradius1,grainclass(1)%refindx*DBLE(1.0),angles,
     & s11Mie1,s12Mie1,CextMie1,CscaMie1)
!
      CextMie1=CextMie1*pi*(mieradius1/wavenumber)**2
      CscaMie1=CscaMie1*pi*(mieradius1/wavenumber)**2
!
      IF (subgrains==1) THEN
         WRITE (outfile,*)'Scattering by a particle with'
         WRITE (outfile,'(A,3F12.4)') '"Size" =',grainclass(1)%a
         WRITE (outfile,'(A,F15.10)')  'Volume =',cvol
         WRITE (outfile,'(A,F12.4,F12.9,A)')  'Refractive index =',
     &                        grainclass(1)%refindx,'i'
         WRITE (outfile,'(A,3F12.4)') 'Wavelength =',wavelength
      ELSE
         WRITE (outfile,*)'Scattering by a cluster of ',subgrains,
     &                   ' subgrains with total volume ',cvol
      ENDIF
      write (outfile,*)
      write (outfile,'(A,F13.8,A)')
     &   'Compared to equivolume sphere of radius',
     &                          mieradius,' (EVS)'
      IF (subgrains>1) write (outfile,'(A,I4,A,F13.8,A)') ' and ',
     &  subgrains,' independent spheres of radius',mieradius1,'(IDS)'
      write (outfile,*) 'Cross section            C           C/V
     &  std.err.    C/C_EVS    C/C_IDS'
      write(outfile,'(A,2ES15.5,A,F7.4,A,2F10.5)') 'Extinction:',Cext,
     &   Cext/cvol,'  +-',
     &   sqrt(max(REAL(0.0),Cext2-Cext**2)/wavenumber)/Cext*100,'%',
     &    Cext/CextMie,Cext/CextMie1/subgrains
      IF (abs(Cabs)<Cext*Epsilon(Cext)*33) THEN
      write (outfile,'(A,2ES15.5,A,F7.4,A,F10.4)') 'Absorption:',Cabs,
     &   Cabs/cvol,'  +-',0.0,'%',0.0
      ELSE
      write(outfile,'(A,2ES15.5,A,F7.4,A,2F10.5)')'Absorption:',Cabs,
     &    Cabs/cvol,'  +-',
     &   sqrt(max(REAL(0.0),Cabs2-Cabs**2)/wavenumber)/Cabs*100,'%',
     &   Cabs/(CextMie-CscaMie),Cabs/(CextMie1-CscaMie1)/subgrains
      ENDIF
      write(outfile,'(A,2ES15.5,A,F7.4,A,2F10.5)') 'Scattering:',Csca,
     &   Csca/cvol,'  +-',
     &   sqrt(max(REAL(0.0),Csca2-Csca**2)/wavenumber)/Csca*100,'%',
     &   Csca/CscaMie,Csca/CscaMie1/subgrains


      IF (abs(Csca-Cext+Cabs)/Cext >reqaccuracy) THEN
        write (outfile,'(A,F15.9,A)') 'SPURIOUS LOSS/GAIN OF ENERGY:',
     &  (Csca-Cext+Cabs)/Cext ,'  !!!!!!!!'
      ELSE
        write (outfile,'(A,F15.9)') 'Spurious loss/gain of energy',
     &  (Csca-Cext+Cabs)/Cext
      ENDIF
      IF (albedo<DBLE(0.8)) THEN
       write(outfile,'(A,F15.9,A,F15.9,A)')'Single scattering albedo',
     & albedo,'  (',apualbedo,')'
      ELSE
       write(outfile,'(A,F15.9,A,F15.9,A)')'Single scattering albedo',
     & apualbedo,'  (',albedo,')'
      ENDIF
!     write(outfile,'(A,ES15.4,A,F10.4)')'Phaseshift cross section',
!    & Cpha,' +-',sqrt(Cpha2-Cpha**2)/Cpha
      write(outfile,'(A,ES15.5,A,F10.4)')'Polarizati cross section',
     & Cpol,' +-',sqrt(max(Cpol2-Cpol**2,REAL(0.0)))
      write(outfile,'(A,ES15.5,A,F10.4)')'Backscattr cross section',
     & Cbak,' +-',sqrt(max(Cbak2-Cbak**2,REAL(0.0)))
      write(outfile,'(A,F11.3)')'Asymmetry parameter<cos>',asymmetry
      write(outfile,*)
      write(outfile,*)'Scat angl   P11 +-..            P21 +-..
     &  P11/P11_EVS   P22-P22_EVS'
      DO i=1,angles,max(1,angles/19)
      write(outfile,'(F5.1,2(F12.5,F8.5),2F11.4)')ScatteringAngle(i),
     &        PhaseMatrix(i,1,1),PhaseMatErr(i,1,1),
     &        PhaseMatrix(i,2,1),PhaseMatErr(i,2,1),
     &        PhaseMatrix(i,1,1)/s11Mie(i),
     &        PhaseMatrix(i,2,1)-s12Mie(i)
!    &        PhaseMatrix(i,3,3),
!    &        PhaseMatrix(i,4,4)
      END DO 
!     close(outfile)
!
!  on standard output
!
      IF (subgrains==1) THEN 
	 WRITE (*,*)'Scattering by a particle with'
	 WRITE (*,'(A,3F12.4)') '"Size" =',grainclass(1)%a
	 WRITE (*,'(A,F15.10)')  'Volume =',cvol
	 WRITE (*,'(A,F12.4,F12.9,A)')  'Refractive index =',
     &                        grainclass(1)%refindx,'i'
	 WRITE (*,'(A,3F12.4)') 'Wavelength =',wavelength
      ELSE
	 WRITE (*,*)'Scattering by a cluster of ',subgrains,
     &                   ' subgrains with total volume ',cvol
      ENDIF
!
      write (*,*) 
      write (*,'(A,F13.8,A)') 
     &   'Compared to equivolume sphere of radius',
     &                          mieradius,' (EVS)'
      IF (subgrains>1) write (*,'(A,I4,A,F13.8,A)') ' and ',subgrains,
     &           ' independent spheres of radius',mieradius1,'(IDS)'
      write (*,*) 'Cross section            C           C/V      
     &  std.err.    C/C_EVS    C/C_IDS'
      write(*,'(A,2ES15.5,A,F7.4,A,2F10.5)') 'Extinction:',Cext,
     &   Cext/cvol,'  +-',
     &   sqrt(max(REAL(0.0),Cext2-Cext**2)/wavenumber)/Cext*100,'%',
     & 	  Cext/CextMie,Cext/CextMie1/subgrains
!
      IF (abs(Cabs)<Cext*Epsilon(Cext)*33) THEN
      write (*,'(A,2ES15.5,A,F7.4,A,F10.4)') 'Absorption:',Cabs,
     &   Cabs/cvol,'  +-',0.0,'%',0.0
      ELSE
      write(*,'(A,2ES15.5,A,F7.4,A,2F10.5)')'Absorption:',Cabs,
     &    Cabs/cvol,'  +-',
     &   sqrt(max(REAL(0.0),Cabs2-Cabs**2)/wavenumber)/Cabs*100,'%',
     &   Cabs/(CextMie-CscaMie),Cabs/(CextMie1-CscaMie1)/subgrains
      ENDIF
!
      write(*,'(A,2ES15.5,A,F7.4,A,2F10.5)') 'Scattering:',Csca,
     &   Csca/cvol,'  +-',
     &   sqrt(max(REAL(0.0),Csca2-Csca**2)/wavenumber)/Csca*100,'%',
     &   Csca/CscaMie,Csca/CscaMie1/subgrains


      IF (abs(Csca-Cext+Cabs)/Cext >reqaccuracy) THEN
	write (*,'(A,F15.9,A)') 'SPURIOUS LOSS/GAIN OF ENERGY:',
     &  (Csca-Cext+Cabs)/Cext ,'  !!!!!!!!'
      ELSE
	write (*,'(A,F15.9)') 'Spurious loss/gain of energy',
     &  (Csca-Cext+Cabs)/Cext 
      ENDIF
!
      IF (albedo<DBLE(0.8)) THEN
	 write (*,'(A,F15.9,A,F15.9,A)') 'Single scattering albedo',
     & albedo,'  (',apualbedo,')'
      ELSE
	 write (*,'(A,F15.9,A,F15.9,A)') 'Single scattering albedo',
     & apualbedo,'  (',albedo,')'
      ENDIF
!
!      write (*,'(A,ES15.4,A,F10.4)') 'Phaseshift cross section',Cpha,
!    & ' +-',sqrt(Cpha2-Cpha**2)/Cpha
      write (*,'(A,ES15.5,A,F10.4)')  'Polarizati cross section',Cpol,
     & ' +-',sqrt(max(Cpol2-Cpol**2,REAL(0.0)))
      write (*,'(A,ES15.5,A,F10.4)')  'Backscattr cross section',Cbak,
     & ' +-',sqrt(max(Cbak2-Cbak**2,REAL(0.0)))
      write (*,'(A,F11.3)') 'Asymmetry parameter<cos>',asymmetry
      write (*,*)
      write (*,*) 'Scat angl   P11 +-..            P21 +-..       P11/
     & P11_EVS   P22-P22_EVS'
!
      DO i=1,angles,max(1,angles/19)
!
      write (*,'(F5.1,2(F12.5,F8.5),2F11.4)')  ScatteringAngle(i),
     &        PhaseMatrix(i,1,1),PhaseMatErr(i,1,1),
     &        PhaseMatrix(i,2,1),PhaseMatErr(i,2,1),
     &        PhaseMatrix(i,1,1)/s11Mie(i),
     &        PhaseMatrix(i,2,1)-s12Mie(i)
!
!    &        PhaseMatrix(i,3,3),
!    &        PhaseMatrix(i,4,4)
      END DO
!
      write(*,*)
      write (*,*) 'Full scattering matrix is written into a file '
!    & outfilename,'" unformatted.'
!
!      DO o=1,91
!        write (*,'(I5,2F12.5)') o,LightCurve(o,1)*fourpi/Csca,
!    & LightCurve(o,2)*fourpi/Csca
!      ENDDO
!
      IF ((Csca-Cext+Cabs)/Cext >10*reqaccuracy) THEN
         write(*,*) 'I''m sorry, 
     &  but these results have quite a low accuracy.'
         write(*,*) 'Please, evaluate the usefulness of 
     &  the results carefully yourself.'
      write(*,*) 'A possible rescue is to repeat computation with 
     &   denser grid.'
      write(*,*) 'In some cases even _lowering_  the expansion 
     &   order may help.'
      ENDIF
      IF ((Csca-Cext+Cabs)/Cext <-10*reqaccuracy) THEN
        write(*,*) 'I''m sorry, but these results 
     &  have quite a low accuracy.'
        write(*,*) 'Please, evaluate the usefulness of 
     &  the results carefully yourself.'
         write(*,*) 'A possible rescue is to repeat computation 
     &  with increased expanion order.'
      ENDIF
!
      CALL DATE_AND_TIME(values=timevalue)
	  write (*,'(A,I2,A,I2,A,I5,A,I2,A,I2,A)')
     &  'Program ended ',timevalue(3),'.',timevalue(2),'.',
     &  timevalue(1),' at ',timevalue(5),':',timevalue(6),' o''clock.'

      write(*,*) 'CPU time total',CPUtime()

      END SUBROUTINE FinalOutput

      END MODULE Resultia
