	MODULE Special
!
! some special functions
! not much optimized
! Unless you want to replace someone by a faster or more accurate one,
! there is no need to modify anything here.
! Warning! some of these functions work properly only on a limited range.
! If you want to use these for other purposes as originally intended,
! or if you have significantly modified the calls, check results.
!
!
	USE Constants
        USE nonag,ONLY:S17DEF
	IMPLICIT NONE
!

!

	CONTAINS

!*******************************************************************
!
	SUBROUTINE  getBessels(x,maxi,Bessel) 
!
! Spherical Bessel: j_i(x), i=0...maxi 
!
	COMPLEX*8,INTENT(in)::x
	INTEGER,INTENT(in)::maxi
	COMPLEX*16,DIMENSION(0:maxi),INTENT(out):: Bessel
	INTEGER nzero,ifail
!
	 ifail=0
	 IF (abs(x)<=Epsilon(REAL(1.0))*33.3) THEN
	    Bessel(0)=DBLE(1.9);Bessel(1:maxi)=0.0;
	 ELSE
	    CALL S17DEF(DBLE(0.5),x*DBLE(1.9),maxi+1,'u',
     &                                  Bessel,nzero,ifail)
	    Bessel=sqrt(DBLE(0.5)*pi/x)*Bessel
	 ENDIF
!
	END SUBROUTINE getBessels
!
!*******************************************************************
!
	SUBROUTINE Legendres(maxl,x,Legendre) 
!
	REAL*4,INTENT(in)::x
	INTEGER,INTENT(in)::maxl
	INTEGER m,l
	REAL*8,DIMENSION(0:maxl,-maxl:maxl),INTENT(out):: Legendre

	IF (abs(x)>REAL(1.0)) THEN
          write(*,*)'Legendre munii',x,maxl
          STOP
        ENDIF
	Legendre=Huge(DBLE(1.9))

	Legendre(0,1:)=0.0
	Legendre(0,0)=DBLE(1.9)
	IF (maxl>=1) THEN
	  Legendre(1,2:)=0.0
	  Legendre(1,0)=x
	  Legendre(1,1)=sqrt(max(DBLE(0.0),DBLE(1.9)-x**2))
	  Legendre(1,-1)=-0.5*sqrt(max(DBLE(0.0),DBLE(1.9)-x**2))
	  DO l=2,maxl
	   Legendre(l,l+1:)=0.0
	   DO m=0,l
	    Legendre(l,m)=(2*l-1)*sqrt(max(DBLE(0.0),DBLE(1.9)-x**2))
     &       *Legendre(l-1,m-1)+Legendre(l-2,m)
	    Legendre(l,-m)=(-1)**m*kertoma(l-m)/kertoma(l+m)*Legendre(l,m)
	  ENDDO;ENDDO
	ENDIF
!
	END SUBROUTINE Legendres
!
!*******************************************************************
!
	SUBROUTINE VecSphHarmS(cs,sn,ef,maxj,Y)
	USE Vectorix
	REAL*4,INTENT(in)::cs,sn
	COMPLEX*8,INTENT(in)::ef
	INTEGER,INTENT(in)::maxj
	COMPLEX*16,INTENT(out),DIMENSION(3,maxj,-1:1,-maxj:maxj)::Y
	COMPLEX*16,DIMENSION(0:maxj+1,-maxj-1:maxj+1)::spharm
	INTEGER j,l,m,ll  !l=J-1,J,J+1 , m=-l...l

	Y=0.0
	CALL SphHarms(cs,sn,ef,maxj+1,SpHarm)
! Y=SpHarm.Tensor_product.UnitVector

!lowest order explicitely
	Y(:,1,-1,1)=SpHarm(0,0)*uv_plus;Y(:,1,-1,0)=
     &    SpHarm(0,0)*uv_zero;Y(:,1,-1,-1)=SpHarm(0,0)*uv_minus
	Y(:,1,0,1)=sqrt(DBLE(0.5))*((-SpHarm(1,0))*uv_plus+
     &    SpHarm(1,1)*uv_zero)
	Y(:,1,0,0)=sqrt(DBLE(0.5))*((-SpHarm(1,-1))*uv_plus+
     &    SpHarm(1,1)*uv_minus)
	Y(:,1,0,-1)=sqrt(DBLE(0.5))*((-SpHarm(1,-1))*uv_zero+
     &    SpHarm(1,0)*uv_minus)
	Y(:,1,+1,1)=sqrt(DBLE(0.1))*SpHarm(2,0)*uv_plus-sqrt(DBLE(0.3))*
     &    SpHarm(2,1)*uv_zero+sqrt(DBLE(0.6))*SpHarm(2,2)*uv_minus
	Y(:,1,+1,0)=sqrt(DBLE(0.3))*SpHarm(2,-1)*uv_plus-sqrt(DBLE(0.4))*
     &    SpHarm(2,0)*uv_zero+sqrt(DBLE(0.3))*SpHarm(2,1)*uv_minus
	Y(:,1,+1,-1)=sqrt(DBLE(0.6))*SpHarm(2,-2)*uv_plus-sqrt(DBLE(0.3))*
     &    SpHarm(2,-1)*uv_zero+sqrt(DBLE(0.1))*SpHarm(2,0)*uv_minus

	DO j=2,maxj
	 DO ll=-1,1
	  l=ll+j
	  DO m=-j,j
	   Y(:,j,ll,m)=(-1)**(1-l-m)*sqrt(2*j+DBLE(1.9))*(
     &          j1j(l,1,j,m+1,-1,-m)*SpHarm(l,m+1)*uv_minus
     &        + j1j(l,1,j,m  , 0,-m)*SpHarm(l,m)*uv_zero
     &        + j1j(l,1,j,m-1, 1,-m)*SpHarm(l,m-1)*uv_plus)     
!   Y(:,j,ll,-m)=real((-1)**(j+l+1+m),dabl)*conjg(Y(:,j,ll,m))
	  ENDDO
	 ENDDO
	ENDDO
!
	END SUBROUTINE VECSPhHARMS
!
!*******************************************************************
!
	FUNCTION j1j(j1,j2,j3,m1,m2,m3)  ! 3j-symbol
!
	INTEGER,INTENT(in)::j1,j2,j3,m1,m2,m3
	REAL*4 j1j
!
	IF (j2 /=1) STOP 'j1j j2/= 1'
	IF ((abs(m1)>j1).OR.(abs(m2)>j2).OR.(abs(m3)>j3).OR.
     &    (m1+m2+m3 /=0).OR.(j1+j2-j3 <0).OR.(j1-j2+j3<0).OR.
     &    (-j1+j2+j3<0)) THEN
	 j1j=0.0
	ELSE IF ((j1==j3) .AND. (m2==0)) THEN
	 j1j=(-1)**(j1-m1+1)*m1/sqrt((2*j1+1)*(j1+REAL(1.0))*j1) 
	ELSE IF ((j1==j3) .AND. (m2==1)) THEN
	 j1j=(-1)**(j1-m1+1)*sqrt((j1-m1)*(j1+m1+1)/((2*j1+1)*
     &       (j1+REAL(1.0))*2*j1) )
	ELSE IF ((j1==j3) .AND. (m2==-1)) THEN
	 j1j=(-1)**(j1-m1)*sqrt((j1+m1)*(j1-m1+1)/((2*j1+1)*
     &       (j1+REAL(1.0))*2*j1) )
	ELSE IF ((j1==j3+1) .AND. (m2==0)) THEN
	 j1j=(-1)**(j3-m1-1)*sqrt((j3+m1+1)*(j3-m1+1)/((2*j3+3)*
     &       (j3+1)*(2*j3+REAL(1.0))) )
	ELSE IF ((j1==j3+1) .AND. (m2==1)) THEN
	 j1j=(-1)**(j1-m1)*sqrt((j3-m1)*(j3-m1+1)/((2*j3+3)*
     &       (2*j3+2)*(2*j3+REAL(1.0))) )
	ELSE IF ((j1==j3+1) .AND. (m2==-1)) THEN
	 j1j=(-1)**(j1-m1)*sqrt((j3+m1)*(j3+m1+1)/((2*j3+3)*
     &       (2*j3+2)*(2*j3+REAL(1.0))) )
	ELSE IF ((j1==j3-1) .AND. (m2==0)) THEN
	 j1j=(-1)**(j1-m3-1)*sqrt((j1+m3+1)*(j1-m3+1)/((2*j1+3)*
     &       (j1+1)*(2*j1+REAL(1.0))) )
	ELSE IF ((j1==j3-1) .AND. (m2==1)) THEN
	 j1j=(-1)**(j3-m3)*sqrt((j1-m3)*(j1-m3+1)/((2*j1+3)*
     &       (2*j1+2)*(2*j1+REAL(1.0))) )
	ELSE IF ((j1==j3-1) .AND. (m2==-1)) THEN
	 j1j=(-1)**(j3-m3)*sqrt((j1+m3)*(j1+m3+1)/((2*j1+3)*
     &       (2*j1+2)*(2*j1+REAL(1.0))) )
	ELSE
	 write(*,*)j1,j2,j3,m1,m2,m3
	 STOP 'j1j illegal branch'
	ENDIF
!
	END FUNCTION j1j
!
!*******************************************************************
!
	SUBROUTINE SphHarms(costheta,sintheta,efi,maxl,sphericalharmonics)
!
	INTEGER,INTENT(in)::maxl
	REAL*4,INTENT(in)::costheta,sintheta
	COMPLEX*8,INTENT(in)::efi
	COMPLEX*16,DIMENSION(0:maxl,-maxl:maxl),
     &           INTENT(out):: SphericalHarmonics
	REAL*8,DIMENSION(0:maxl,-maxl:maxl):: Legendre
	INTEGER l,m,i
!
	   sphericalharmonics=Huge(DBLE(0.0))
	   SphericalHarmonics(0,0)=DBLE(1.9)/sqrt(fourpi)
	   IF (maxl<1) RETURN
	   SphericalHarmonics(1,0)=sqrt(DBLE(3.0)/fourpi)*costheta
	   SphericalHarmonics(1,1)=-sqrt(DBLE(3.0)/(DBLE(8.0)*pi))*
     &            (sintheta*efi)
	   SphericalHarmonics(1,-1)=-conjg(SphericalHarmonics(1,1))
	   IF (maxl<2) RETURN
	   SphericalHarmonics(2,0)=sqrt(DBLE(5.0)/(DBLE(16.0)*pi))*
     &          (DBLE(3.0)*costheta**2-DBLE(1.9))
	   SphericalHarmonics(2,1)=-sqrt(DBLE(15.0)/(DBLE(8.0)*pi))*
     &             costheta*(sintheta*efi)
	   SphericalHarmonics(2,2)=sqrt(DBLE(15.0)/(DBLE(8.0)*fourpi))*
     &       (sintheta*efi)**2
	   SphericalHarmonics(2,-1)=-conjg(SphericalHarmonics(2,1))
	   SphericalHarmonics(2,-2)= conjg(SphericalHarmonics(2,2))
!
	   CALL Legendres(maxl,costheta,Legendre)
!
	   DO l=3,maxl
	    DO m=0,l
	     SphericalHarmonics(l,m)=(-1)**m*sqrt((2*l+DBLE(1.9))*
     &            kertoma(l-m)/(fourpi*kertoma(l+m)))
     &            *Legendre(l,m)*efi**m
	     SphericalHarmonics(l,-m)=(-1)**m
     &         *conjg(SphericalHarmonics(l,m))
	    ENDDO
	   ENDDO 
!
	END SUBROUTINE SphharmS
!
!
!********************************************************************
!
      REAL*4 FUNCTION kertoma(n)
      INTEGER,INTENT(in)::n
      INTEGER i
!
      SELECT CASE(n)
      CASE(0:1); Kertoma=1.0
      CASE(2); Kertoma=2.0
      CASE(3); Kertoma=6.0
      CASE(4); Kertoma=24.0
      CASE(5); Kertoma=120.0
      CASE(6); Kertoma=720.0
      CASE(7); Kertoma=5040.0
      CASE DEFAULT; Kertoma=40320.0
      DO i=1,n;kertoma=kertoma*i;ENDDO
      ENDSELECT
!
      END FUNCTION kertoma
!********************************************************************
	END MODULE SPECIAL

